#include "defs.h"

#include <linux/input.h>
#include <linux/uinput.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>


#define UINPUT_DEVICE "/dev/input/uinput"

static int device = -1;
static int created = 0;
static struct uinput_user_dev uinput = { };


#define WRITE_EVENT(_type, _code, _value) \
	event.type = (_type); \
	event.code = (_code); \
	event.value = (_value); \
	gettimeofday(&(event.time), NULL); \
\
	res = TEMP_FAILURE_RETRY(write(device, &event, sizeof(event))); \
	if (res != sizeof(event)) { \
		perror("can not inject event"); \
		return -errno; \
	}

static int write_data(struct output_data *out)
{
	struct input_event event;
	int res;

	if (device < 0)
		return -ENODEV;

	WRITE_EVENT(EV_REL, REL_X, out->X);
	WRITE_EVENT(EV_REL, REL_Y, out->Y);

	WRITE_EVENT(EV_KEY, BTN_LEFT, out->buttons.left);
	WRITE_EVENT(EV_KEY, BTN_RIGHT, out->buttons.right);
	WRITE_EVENT(EV_KEY, BTN_MIDDLE, out->buttons.middle);

	WRITE_EVENT(EV_SYN, SYN_REPORT, 0);

	return 0;
}

static void cleanup(void)
{
	int res;

	if (device < 0)
		return;

	if (created) {
		res = TEMP_FAILURE_RETRY(ioctl(device, UI_DEV_DESTROY, 0));
		if (res < 0) {
			perror("can not create input device");
		}
	}

	res = TEMP_FAILURE_RETRY(close(device));
	if (res < 0)
		perror("can not close uinput device");
}

#define SET_UI_BIT(_type, _code) \
res = TEMP_FAILURE_RETRY(ioctl(device, (_type), (_code))); \
if (res < 0) { \
	perror("can not set input bit"); \
	return -errno; \
}

static int init_device(void)
{
	int res;

	res = TEMP_FAILURE_RETRY(open(UINPUT_DEVICE, O_RDWR));
	if (res < 0) {
		perror("can not open uinput device");
		return -errno;
	}
	device = res;

	SET_UI_BIT(UI_SET_EVBIT, EV_REL);
	SET_UI_BIT(UI_SET_RELBIT, REL_X);
	SET_UI_BIT(UI_SET_RELBIT, REL_Y);

	SET_UI_BIT(UI_SET_EVBIT, EV_KEY);
	SET_UI_BIT(UI_SET_KEYBIT, BTN_LEFT);
	SET_UI_BIT(UI_SET_KEYBIT, BTN_RIGHT);
	SET_UI_BIT(UI_SET_KEYBIT, BTN_MIDDLE);

	snprintf(uinput.name, UINPUT_MAX_NAME_SIZE, "stick2mouse");
	uinput.id.bustype = BUS_USB;
	uinput.id.vendor = 0x06cb;
	uinput.id.product = 0x0009;
	uinput.id.version = 0;

	res = TEMP_FAILURE_RETRY(write(device, &uinput, sizeof(uinput)));
	if (res != sizeof(uinput)) {
		perror("can not write uinput setup data");
		return -errno;
	}

	res = TEMP_FAILURE_RETRY(ioctl(device, UI_DEV_CREATE, 0));
	if (res < 0) {
		perror("can not create input device");
		return -errno;
	}
	created = 1;

	return 0;
}

struct output_functions output_uinput = {
	.init = init_device,
	.exit = cleanup,
	.put_data = write_data
};
